import logging

from .constants import *

logger = logging.getLogger(__name__)

# Device profiles
class DeviceProfile:
    def __init__(self, name: str, address: int, device_id: int, capabilities: list[int],
                 typ: int, interface: int = ADAFRUIT):
        self._name = name
        self._address = address
        self._dev_id = device_id
        self._caps = capabilities
        self._type = typ
    
    @property
    def address(self) -> int:
        return self._address

    @property
    def capabilities(self) -> list[int]:
        return self._caps

    @property
    def name(self) -> str:
        return self._name

    @property
    def device_id(self) -> int:
        return self._dev_id

    @property
    def typ(self) -> int:
        return self._type

    def is_mux(self) -> bool:
        return True if self._type == MUX else False

    def is_meter(self) -> bool:
        return True if self._type == METER else False

    def is_detector(self) -> bool:
        return True if self._type == DETECTOR else False

    def has_capability(self, cap: int) -> bool:
        return True if cap in self._caps else False

pca9546a = DeviceProfile('PCA9546A', 0x70, PCA9546A, [ FOUR_CHANNEL ], MUX)
tca9548a = DeviceProfile('TCA9548A', 0x70, TCA9548A, [ EIGHT_CHANNEL ], MUX)
bmp390   = DeviceProfile('BMP390', 0x77, BMP390, [ PRESSURE, TEMPERATURE, ALTITUDE ], METER)
sht41    = DeviceProfile('SHT41', 0x44, SHT41, [ TEMPERATURE, RELATIVE_HUMIDITY ], METER)
veml7700 = DeviceProfile('VEML7700', 0x10, VEML7700, [ AMBIENT_LIGHT, LUX ], METER)
scd41    = DeviceProfile('SCD41', 0x62, SCD41, [ CO2, RELATIVE_HUMIDITY, TEMPERATURE ], METER)
tsl2591  = DeviceProfile('TSL2591', 0x29, TSL2591,
                         [ LUX, FULL_SPECTRUM, VISIBLE, INFRARED, AMBIENT_LIGHT ], METER)

profiles = dict([
    (0x70, pca9546a),
    (0x77, bmp390),
    (0x44, sht41),
    (0x10, veml7700),
    (0x62, scd41),
    (0x29, tsl2591),
])



import logging
logger = logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
        'calibration',
        'config',
        'constants',
        'controls',
        'datastructures',
        'devices',
        'devicetree',
        'meters',
        'profiles',
        'virtuals',
]

from .sensorkit import SensorKit
from . import calibration
from . import config
from . import constants
from . import controls
from . import datastructures
from . import devices
from . import devicetree
from . import meters
from . import profiles
from . import virtuals



import abc
from collections.abc import Iterator
import logging
from typing import Literal

logger = logging.getLogger(__name__)

import board
from busio import I2C
import adafruit_tca9548a

from . import constants
from . import devices

class MuxInterface(devices.DeviceInterface, metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'address') and
                callable(subclass.address) and
                hasattr(subclass, 'channels') and
                callable(subclass.channels) and
                hasattr(subclass, '__len__') and
                callable(subclass.__len__) or
                NotImplemented)

    @abc.abstractmethod
    def channels(self) -> Iterator[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        raise NotImplementedError

class PCA9546A(devices.Device, MuxInterface):
    def __init__(self, bus: I2C, name: str, board: int, capabilities: list[int],
                 address: int | str = 112):
        super().__init__(bus, name, board, capabilities, address)
        self._mux = adafruit_tca9548a.PCA9546A(bus, address)
        pass

    def __len__(self) -> Literal[4]:
        return len(self._mux)

    @property
    def real_device(self):
        return self._mux

    def channels(self) -> Iterator[int]:
        for c in range(len(self._mux)):
            yield self._mux[c]

class MuxFactory:
    def __init__(self):
        self._ctors = {}

    def register_mux(self, board, ctor):
        self._ctors[board] = ctor

    def get_mux(self, bus: I2C, name: str, board: int, capabilities: list[int],
                address: int | str) -> MuxInterface:
        ctor = self._ctors.get(board)
        if not ctor:
            raise ValueError(board)

        return ctor(bus, name, board, capabilities, address)

mux_factory = MuxFactory()
mux_factory.register_mux(constants.PCA9546A, PCA9546A)



NONE = 0x00

# Addresses
VIRTUAL_ADDR = 0xFF

# Interfaces
ADAFRUIT = 0x0001
DFROBOT  = 0x0002

# Device interfaces
adafruit = 'adafruit'
dfrobot  = 'dfrobot'

to_interface = dict({
    adafruit: ADAFRUIT,
    dfrobot:  DFROBOT
})

# Device Types
VIRTUAL       = 0x80000000

BUS           = 0x0001
MUX           = 0x0002
CHANNEL       = 0x0004
DEVICE        = 0x0008
METER         = 0x0010
DETECTOR      = 0x0020

# Type strings
mux      = 'mux'
meter    = 'meter'
detector = 'detector'

to_device_type = dict({
    mux:      MUX,
    meter:    METER,
    detector: DETECTOR
})

# Capabilities
FOUR_CHANNEL            = 0x0001
EIGHT_CHANNEL           = 0x0002
PRESSURE                = 0x0004
TEMPERATURE             = 0x0008
ALTITUDE                = 0x0010
RELATIVE_HUMIDITY       = 0x0020
AMBIENT_LIGHT           = 0x0040
LUX                     = 0x0080
VISIBLE                 = 0x0100
INFRARED                = 0x0200
FULL_SPECTRUM           = 0x0400
CO2                     = 0x0800
PRESSURE_MSL            = 0x1000

# Capability Strings
four_channel      = 'four_channel'
eight_channel     = 'eight_channel'
pressure          = 'pressure'
temperature       = 'temperature'
altitude          = 'altitude'
relative_humidity = 'relative_humidity'
ambient_light     = 'ambient_light'
lux               = 'lux'
visible           = 'visible'
infrared          = 'infrared'
full_spectrum     = 'full_spectrum'
co2               = 'CO2'
pressure_msl      = 'pressure_msl'

to_capability_strings = dict([
    (FOUR_CHANNEL,            four_channel),
    (EIGHT_CHANNEL,           eight_channel),
    (PRESSURE,                pressure),
    (TEMPERATURE,             temperature),
    (ALTITUDE,                altitude),
    (RELATIVE_HUMIDITY,       relative_humidity),
    (AMBIENT_LIGHT,           ambient_light),
    (LUX,                     lux),
    (VISIBLE,                 visible),
    (INFRARED,                infrared), 
    (FULL_SPECTRUM,           full_spectrum),
    (CO2,                     co2),
    (PRESSURE_MSL,            pressure_msl),
])

to_capabilities = {v: k for k, v in to_capability_strings.items()}

# Units
CELSIUS_UNITS                = 'Celsius (C)'
HECTOPASCAL_UNITS            = 'Hectopascal (hPa)'
METER_UNITS                  = 'Meter (M)'
PERC_RELATIVE_HUMIDITY_UNITS = 'Percent Relative Humidity (% rH)'
AMBIENT_LIGHT_UNITS          = 'Ambient Light Data'
LUX_UNITS                    = 'Lux (Lx)'
PPM_UNITS                    = 'Parts per Million (PPM)'

# Breakout boards
VIRTUAL_DEVICE = NONE
PCA9546A       = 0x0001
PCA9548A       = 0x0002
TCA9548A       = PCA9548A
BMP390         = 0x0003
SHT41          = 0x0004
VEML7700       = 0x0005
SCD41          = 0x0006
TSL2591        = 0x0007

to_device_ids = dict({
    'virtual':  VIRTUAL_DEVICE,
    'pca9546a': PCA9546A,
    'pca9548a': PCA9548A,
    'tca9548a': TCA9548A,
    'bmp390':   BMP390,
    'sht41':    SHT41,
    'veml7700': VEML7700,
    'scd41':    SCD41,
    'tsl2591':  TSL2591,
})



import abc
import json
import logging

from isodate import parse_duration

from ..constants import (VIRTUAL_DEVICE,
                         PRESSURE_MSL,
                         TEMPERATURE,
                         RELATIVE_HUMIDITY,
                         to_capabilities,
)
from ..datastructures import Store
from ..devices import VirtualDevice
from ..meters import Meter
from ..tools.mixins import GetterMixin, SchedulableMixin

logger = logging.getLogger(__name__)

to_open_meteo = dict({
    'temperature': 'temperature_2m',
    'relative_humidity': 'relative_humidity_2m',
})
from_open_meteo = { v:k for k,v in zip(to_open_meteo.keys(), to_open_meteo.values()) }

open_meteo_measures = dict({
    'temperature_2m': TEMPERATURE,
    'relative_humidity_2m': RELATIVE_HUMIDITY,
})
open_meteo_measures = { **open_meteo_measures, **to_capabilities }

class OpenMeteoMixin_(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, '_handler') and
                callable(subclass._handler) or
                NotImplemented)

    @abc.abstractmethod
    def _handler(self) -> None:
        raise NotImplementedError

class OpenMeteoCurrentGetterImpl_(GetterMixin, SchedulableMixin):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super(OpenMeteoCurrentGetterImpl_, cls).__new__(cls)
        return cls._instance
    
    def __init__(self, capabilities: list[str], interval: str, params: dict[str, int | str],
                 store: Store, scheduler):
        self._location = 'https://api.open-meteo.com/v1/forecast?'
        self._handlers = {}

        self._interval = parse_duration(interval).seconds
        self._params = params
        self._capabilities = capabilities
        self._scheduler = scheduler
        self._job = None

        param = [ c if c not in to_open_meteo else to_open_meteo[c] for c in capabilities]
        self._params['current'] = ','.join(param)

    def schedule(self, immediate: bool) -> None:
        if self._job is not None:
            return

        if immediate:
            self.url_get(self._params)

        self._job = self._scheduler.add_job(self.url_get, 'interval', seconds=self._interval,
                                            kwargs = { 'params': self._params })

    def unschedule(self) -> None:
        if self._job is None:
            return

        self._job.remove()
        self._job = None

    def set_handler(self, capability: str, handler: callable) -> None:
        if capability not in self._capabilities:
            logger.warning('open meteo impl: requested unset cap: %s', capability)
            return

        self._handlers[capability] = handler

    @property
    def location(self) -> str:
        return self._location

    def _handler(self, contents) -> None:
        if contents.status != 200:
            logger.warning('open meteo impl: API failed GET - %s', contents.status)
            return

        data = contents.read()
        obj = json.loads(data)

        current_values = obj['current']
        current_units = obj['current_units']

        for c in self._handlers:
            func = self._handlers[c]
            v = current_values[c] if c not in to_open_meteo else current_values[to_open_meteo[c]]
            u = current_units[c] if c not in to_open_meteo else current_units[to_open_meteo[c]]
            func(c, v, u)

class OpenMeteoCurrent_(Meter, OpenMeteoMixin_):
    def __init__(self, capability: str, interval: str,
                 params: dict[str, int | str], store: Store,
                 scheduler, device: VirtualDevice | None = None):
        if device is None:
            super().__init__(VirtualDevice(None, 'open-meteo', VIRTUAL_DEVICE,
                                           [ open_meteo_measures[capability] ]))
        else:
            super().__init__(device)

        self._capability = capability
        self._store = store

    @property
    def measure(self) -> float:
        return getattr(self._store, self._capability)

    @property
    def measurement(self) -> int:
        return open_meteo_measures[self._capability] 

    @property
    def units(self) -> str:
        units_key = self._capability + '_units'
        return getattr(self._store, units_key)

    @property
    def real_device(self):
        return None

    def _handler(self, capability: str, value: float | int, units: str) -> None:
        logger.debug('open_meteo_handler called with %s %s %s',
                     capability, value, units)

        if capability != self._capability:
            msg = 'open meteo meter - called with mismatch capability: got %s != wanted %s'
            logger.warning(msg, capability, self._capability)
            return

        setattr(self._store, capability, value)
        setattr(self._store, capability + '_units', units)

class OpenMeteoCurrentBuilder:
    def __init__(self, capabilities: list[str]):
        self._caps = capabilities

    def __call__(self, interval: str, params: dict[str, int | str], store: Store,
                 scheduler, **_ignored) -> list[OpenMeteoCurrent_ | OpenMeteoCurrentGetterImpl_]:
        # the builder knows what this is, put your smarts here when building other layered
        # virtual devices

        getter_impl = OpenMeteoCurrentGetterImpl_(self._caps, interval, params, store, scheduler)

        devs = []
        for cap in self._caps:
            obj = OpenMeteoCurrent_(cap, interval, params, store, scheduler)
            getter_impl.set_handler(cap, obj._handler)
            devs.append(obj)

        getter_impl.schedule(True)

        return devs



import abc
from collections.abc import Iterator
import logging
from typing import Any

import adafruit_bmp3xx
import adafruit_scd4x
import adafruit_sht4x
import adafruit_tsl2591
import adafruit_veml7700
from busio import I2C

from . import constants
from . import profiles
from .tools.mixins import RunnableMixin

logger = logging.getLogger(__name__)

class DeviceInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'address') and
                callable(subclass.address) and
                hasattr(subclass, 'board') and
                callable(subclass.board) and
                hasattr(subclass, 'name') and
                callable(subclass.name) and
                hasattr(subclass, 'capabilities') and
                callable(subclass.capabilities) and
                hasattr(subclass, 'capabilities_gen') and
                callable(subclass.capabilities_gen) and
                hasattr(subclass, 'bus') and
                callable(subclass.bus) and
                hasattr(subclass, 'bus_id') and
                callable(subclass.bus_id) and
                hasattr(subclass, 'real_device') and
                callable(subclass.real_device) or
                NotImplemented)

    @abc.abstractmethod
    def address(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def board(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def capabilities(self) -> list[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def capabilities_gen(self) -> Iterator[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def bus(self) -> I2C:
        raise NotImplementedError

    @abc.abstractmethod
    def bus_id(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def real_device(self):
        raise NotImplementedError

class Device(DeviceInterface):
    def __init__(self, bus: I2C | None, name: str, board: int, capabilities: list[int],
                 address: int, env: dict[str, Any] | None = None):
        self._bus = bus
        self._name = name
        self._board = board
        self._caps = capabilities
        self._address = address

    @property
    def address(self) -> int | str:
        return self._address

    @property
    def board(self) -> int:
        return self._board

    @property
    def name(self) -> str:
        return self._name

    @property
    def capabilities(self) -> list[int]:
        return self._caps

    def capabilities_gen(self) -> Iterator[int]:
        for cap in self._caps:
            yield cap

    @property
    def bus(self) -> I2C:
        return self._bus

    @property
    def bus_id(self) -> int:
        r = 0
        if self._bus is not None and 'channel_switch' in self._bus.__dict__:
            n = int.from_bytes(self._bus.channel_switch, 'little')
            while n > 1:
                n = n >> 1
                r = r + 1
        return r

class VirtualDevice(Device):
    def __init__(self, name: str, board: int, capabilities: list[int],
                 address: int = constants.VIRTUAL_ADDR, env: dict[str, Any] | None = None):
        super().__init__(None, name, board, capabilities, address)
    
    @property
    def real_device(self):
        return None

class Bmp390(Device):
    def __init__(self, bus: I2C, name: str, board: int, capabilities: list[int],
                 address: int = 119, env: dict[str, Any] | None = None):
        super().__init__(bus, name, board, capabilities, address)
        self._dev = adafruit_bmp3xx.BMP3XX_I2C(bus, address)

    @property
    def real_device(self):
        return self._dev

class Sht41(Device):
    def __init__(self, bus: I2C, name: str, board: int, capabilities: list[int],
                 address: int = 68, env: dict[str, Any] | None = None):
        super().__init__(bus, name, board, capabilities, address)
        self._dev = adafruit_sht4x.SHT4x(bus, address)

    @property
    def real_device(self):
        return self._dev

class Veml7700(Device):
    def __init__(self, bus: I2C, name: str, board: int, capabilities: list[int],
                 address: int = 16, env: dict[str, Any] | None = None):
        super().__init__(bus, name, board, capabilities, address)
        self._dev = adafruit_veml7700.VEML7700(bus, address)

    @property
    def real_device(self):
        return self._dev

class Scd41(Device, RunnableMixin):
    def __init__(self, bus: I2C, name: str, board: int, capabilities: list[int],
                 address: int = 98, env: dict[str, Any] | None = None):
        super().__init__(bus, name, board, capabilities, address)
        self._dev = adafruit_scd4x.SCD4X(bus, address)
        self._dev.reinit()

        self._env = env

    @property
    def real_device(self):
        return self._dev

    def run(self):
        if 'indoors' in self._env:
            enable = not self._env['indoors']
            self._dev.self_calibration_enabled = enable

        self._dev.start_periodic_measurement()

    def stop(self):
        self._dev.stop_periodic_measurement()

class Tsl2591(Device):
    def __init__(self, bus: I2C, name: str, board: int, capabilities: list[int],
                 address: int = 41, env: dict[str, Any] | None = None):
        super().__init__(bus, name, board, capabilities, address)
        self._dev = adafruit_tsl2591.TSL2591(bus, address)

    @property
    def real_device(self):
        return self._dev

class DeviceFactory:
    def __init__(self):
        self._ctors = {}

    def register_device(self, board, ctor):
        self._ctors[board] = ctor

    def get_device(self, bus: I2C, name: str, board: int, capabilities: list[int],
                   address: int, env: dict[str, Any] | None = None) -> DeviceInterface:
        ctor = self._ctors.get(board)
        if not ctor:
            raise ValueError(board)

        return ctor(bus, name, board, capabilities, address, env)

device_factory = DeviceFactory()
device_factory.register_device(constants.BMP390, Bmp390)
device_factory.register_device(constants.SHT41, Sht41)
device_factory.register_device(constants.VEML7700, Veml7700)
device_factory.register_device(constants.SCD41, Scd41)
device_factory.register_device(constants.TSL2591, Tsl2591)



import builtins
import logging

from isodate import parse_duration

from .constants import (
        to_capabilities,
        to_device_ids,
)
from .tools.mixins import SchedulableMixin

logger = logging.getLogger(__name__)

class Calibration(SchedulableMixin):
    def __init__(self, target, conf_dict, target_obj, store, scheduler):
        self._last = None
        self._policy = None
        self._policy_interval = None
        self._target_obj = None
        self._static = None
        self._type = None

        self._target = target
        self._conf = conf_dict
        self._store = store
        self._scheduler = scheduler
        self._job = None

        if self._conf['target']['where'] == 'real':
            self._target_obj = target_obj.real_device
        elif self._conf['target']['where'] == 'abstract':
            self._target_obj = target_obj
        else:
            raise NotImplementedError

        self._by_property = True if 'property' in self._conf['target'] else False
        self._attribute = self._conf['target']['property'] if self._by_property is True \
                else self._conf['target']['method']

        measurement = self._conf['measurement']
        source = self._conf['source']
        func = None
        if 'meter' in source:
            def __filter(node):
                meta = getattr(node, 'metadata', None)
                if meta is not None and meta.is_meter and to_device_ids[source['meter']] == node.obj.board:
                    if to_capabilities[measurement] == node.obj.measurement:
                        return True
                return False
            func = __filter
        elif 'virtual' in source:
            def __filter(node):
                meta = getattr(node, 'metadata', None)
                if meta is not None and meta.is_virtual and source['virtual'] == node.name:
                    if to_capabilities[measurement] == node.obj.measurement:
                        return True
                return False
            func = __filter
        else:
            def __filter(node):
                meta = getattr(node, 'metadata', None)
                if meta is not None and meta.is_meter:
                    if to_capabilities[measurement] == node.obj.measurement:
                        return True
                return False
            func = __filter
        self._sources = self._store.tree.findall(func)

        if self._conf['policy']['aggregation'] == 'average':
            self._policy = self._cast_measure(self._measure_average)
        elif self._conf['policy']['aggregation'] == 'first':
            self._policy = self._cast_measure(self._measure_first)
        else:
            raise NotImplementedError

        if 'type' in self._conf['policy']:
            self._type = self._conf['policy']['type']

        self._policy_interval = 'oneshot' if self._conf['policy']['interval'] == 'oneshot' else \
                parse_duration(self._conf['policy']['interval']).seconds

        logger.debug('config: interval %s policy %s object %s',
                     self._policy_interval, self._policy, self._target_obj)

    def schedule(self, immediate: bool) -> None:
        if self._policy_interval == 'oneshot':
            self._job = 'finished'
            self.calibrate()
            return
        elif self._job is not None:
            return

        if immediate:
            self.calibrate()
        self._job = self._scheduler.add_job(self.calibrate, 'interval',
                                            seconds=self._policy_interval)

    def unschedule(self):
        if self._job is None or self._job == 'finished':
            return

        self._job.remove()
        self._job = None

    @property
    def static(self):
        return self._static

    def _cast_measure(self, func: callable):
        def __cast():
            v = func()
            v = v if self._type is None else getattr(builtins, self._type)(v)
            return v
        return __cast

    def _measure_average(self):
        l = len(self._sources)
        v = 0
        for s in self._sources:
            v = v + s.measure
        v = v / l
        return v

    def _measure_first(self):
        s = self._sources[0]
        return s.measure

    def _measure_specific(self):
        raise NotImplementedError

    def _setter(self, value):
        _value = value if self._type is None else getattr(builtins, self._type)(value)
        if self._by_property:
            setattr(self._target_obj, self._attribute, _value)
        else:
            func = getattr(self._target_obj, self._attribute)
            func(_value)
        self._last = _value

    def calibrate(self):
        logger.debug('Calibration.calibrate source %s setting %s for %s',
                     self._conf['source'], self._attribute,
                     self._target)
        value = self._policy()
        if self._last is None or self._last != value:
            logger.info('found change from %s to %s', self._last, value)
            logger.debug('will pull new value %s from %s.measure', value,
                         self._conf['source'])
            logger.debug('will set value %s to %s.%s with %s device', value, self._target,
                         self._attribute, self._conf['target']['where'])

            self._setter(value)
        else:
            logger.debug('no changes in %s', self._attribute)



from typing import Any
import logging
import os.path

import yaml

logger = logging.getLogger(__name__)

class Config:
    def __init__(self, config: dict[str, Any]):
        self._data = config

        self._metrics_endpoint = None
        self._metrics_encoding = None
        self._metrics_labels = None

        self._virtual_devices = None
        self._calibrations = None
        self._env = None

    @property
    def env(self) -> dict[str, Any]:
        if self._env is not None:
            return self._env

        self._env = self._data.get('env', {})
        return self._env

    @property
    def virtual_devices(self) -> dict[str, dict[str, str]]:
        if self._virtual_devices is not None:
            return self._virtual_devices

        self._virtual_devices = self._data.get('virtual-devices', {})
        return self._virtual_devices

    @property
    def indoors(self):
        if self._indoors is not None:
            return self._indoors

        self._indoors = self._data.get('indoors', True)
        return self._indoors

    @property
    def calibrations(self) -> dict[str, dict[str, str]]:
        if self._calibrations is not None:
            return self._calibrations

        self._calibrations = self._data.get('calibrations', {})
        return self._calibrations



import logging
import typing

logger = logging.getLogger(__name__)

class Store:
    _store: dict[str, typing.Any]
    _listeners: dict[str, list[callable]]

    def __init__(self, state: dict[str, typing.Any] | None = None):
        if state is None:
            state = {}
        super().__setattr__("_store", state)
        super().__setattr__("_listeners", {})

    def __setattr__(self, key: typing.Any, value: typing.Any) -> None:
        self._store.__setattr__(key, value)

        if key in self._listeners:
            for h in self._listeners[key]:
                h(key, value)

    def __getattr__(self, key: typing.Any) -> typing.Any:
        try:
            return self._store.__getattr__(key)
        except (KeyError, AttributeError):
            message = "'{}' object has no attribute '{}'"
            raise AttributeError(message.format(self.__class__.__name__, key))

    def __delattr__(self, key: typing.Any) -> None:
        del self._store[key]

        if key in self._listeners:
            for h in self._listeners[key]:
                h(key, None)

    def add_key_listener(self, key: typing.Any, handler: callable) -> None:
        if key in self._listeners:
            self._listeners[key].append(handler)
        else:
            self._listeners[key] = [ handler ]



import logging
from typing import Any

from ..constants import (VIRTUAL_DEVICE,
                         to_capabilities,
)
from ..datastructures import Store
from ..devices import VirtualDevice
from ..meters import Meter

logger = logging.getLogger(__name__)

class StaticDevice(Meter):
    def __init__(self, capability: str, value: Any, units: str):
        super().__init__(VirtualDevice(None, 'static-virtual', VIRTUAL_DEVICE,
                                       [ to_capabilities[capability] ]))
        self._capability = capability
        self._value = value
        self._units = units

    @property
    def measure(self) -> Any:
        return self._value

    @property
    def measurement(self) -> int:
        return to_capabilities[self._capability]
    
    @property
    def units(self) -> str:
        return self._units

    @property
    def real_device(self):
        return None

class StaticBuilder:
    def __init__(self, capabilities: list[str]):
        self._caps = capabilities

    def __call__(self, values: dict[str, dict[str, Any]], store: Store, **_ignored):
        devs = []
        for cap in self._caps:
            if cap not in values:
                raise ValueError('need value for capability {}'.format(cap))

            value = values[cap]
            if 'value' not in value or 'units' not in value:
                raise ValueError('need value and units for capability {}'.format(cap))


            obj = StaticDevice(cap, value['value'], value['units'])
            devs.append(obj)

        return devs



import logging
from collections.abc import Iterator
from typing import Any

from anytree import Node, PreOrderIter
from anytree.search import find
import board
from busio import I2C

from . import constants
from . import controls
from . import datastructures
from . import devices
from . import meters
from . import profiles

logger = logging.getLogger(__name__)

ROOT = constants.NONE

ROOT_NAME    = 'root'
VIRTUAL_NAME = 'virtual-bus'
I2C_NAME     = 'i2c-bus'

class Metadata:
    def __init__(self, device_type: int):
        self._device_type = device_type

    @property
    def is_root(self) -> bool:
        return self._device_type == constants.NONE

    @property
    def is_physical(self) -> bool:
        return not bool(self._device_type & constants.VIRTUAL)

    @property
    def is_virtual(self) -> bool:
        return bool(self._device_type & constants.VIRTUAL)
        
    @property
    def is_bus(self) -> bool:
        return bool(self._device_type & constants.BUS)

    @property
    def is_virtual_bus(self) -> bool:
        typ = self._device_type
        return bool(typ & constants.BUS) and bool(typ & constants.VIRTUAL)

    @property
    def is_mux(self) -> bool:
        return bool(self._device_type & constants.MUX)

    @property
    def is_channel(self) -> bool:
        return bool(self._device_type & constants.CHANNEL)

    @property
    def is_device(self) -> bool:
        return bool(self._device_type & constants.DEVICE)

    @property
    def is_virtual_device(self) -> bool:
        typ = self._device_type
        return bool(typ & constants.DEVICE) and bool(typ & constants.VIRTUAL)

    @property
    def is_meter(self) -> bool:
        return bool(self._device_type & constants.METER)

    @property
    def is_virtual_meter(self) -> bool:
        typ = self._device_type
        return bool(typ & constants.METER) and bool(typ & constants.VIRTUAL)

    @property
    def is_detector(self) -> bool:
        return bool(self._device_type & constants.DETECTOR)

class DeviceTree:
    def __init__(self, i2c: I2C, store: datastructures.Store, env: dict[str, Any] | None = None):
        self._i2c = i2c
        self._store = store
        self._env = env

        self._root = Node(ROOT_NAME, obj=None, metadata=Metadata(ROOT))
        self._i2c_bus = Node(I2C_NAME, parent=self._root, obj=self._i2c,
                             metadata=Metadata(constants.BUS))
        # /virtual-bus/
        #             |- {name}
        #             |- {name}
        self._virtual_bus = Node(VIRTUAL_NAME, parent=self._root, obj=None,
                                 metadata=Metadata(constants.BUS | constants.VIRTUAL))

    def build(self) -> None:
        self._build_tree(self._i2c, self._i2c_bus, self._store, self._env)

    # Helper to find meters, common use case
    def meters_iter(self, _filter = None) -> Iterator[meters.MeterInterface]:
        if _filter is not None and not callable(_filter):
            raise TypeError('_filter argument must be callable or None')

        def __decorate_filter(_filter = None):
            def __filter(node: Node) -> bool:
                meta = getattr(node, 'metadata', None)
                if meta is not None:
                    is_returnable = meta.is_meter
                    if is_returnable and _filter is not None and callable(_filter):
                        is_returnable = is_returnable and _filter(node)
                    return is_returnable
                return False
            return __filter

        for node in PreOrderIter(self._root, __decorate_filter(_filter)):
            yield node.obj

    def measurement_iter(self, measurement: int) -> Iterator[meters.MeterInterface]:
        def __filter(node: Node) -> bool:
            if hasattr(node, 'obj'):
                obj = node.obj
                if measurement == obj.measurement:
                    return True
            return False

        for m in self.meters_iter(__filter):
            yield m

    def measurement_by_board_iter(self, measurement: int,
                                  board: int) -> Iterator[meters.MeterInterface]:
        def __filter(node: Node) -> bool:
            if hasattr(node, 'obj'):
                obj = node.obj
                if measurement == obj.measurement and board == obj.board:
                    return True
            return False

        for m in self.meters_iter(__filter):
            yield m

    def devices_iter(self, _filter = None) -> Iterator[devices.DeviceInterface]:
        if _filter is not None and not callable(_filter):
            raise TypeError('_filter argument must be callable or None')

        def __decorate_filter(_filter = None):
            def __filter(node: Node) -> bool:
                meta = getattr(node, 'metadata', None)
                if meta is not None:
                    is_returnable = meta.is_device
                    if is_returnable and _filter is not None and callable(_filter):
                        is_returnable = is_returnable and _filter(node)
                    return is_returnable
                return False
            return __filter

        for node in PreOrderIter(self._root, __decorate_filter(_filter)):
            yield node.obj

    def devices_by_board_iter(self, board: int) -> Iterator[devices.DeviceInterface]:
        def __filter(node: Node) -> bool:
            if hasattr(node, 'obj'):
                obj = node.obj
                if board == obj.board:
                    return True
            return False

        for d in self.devices_iter(__filter):
            yield d

    # General purpose filtered iterator
    def filtered_iter(self, _filter: callable) -> Iterator[Any]:
        def __decorate_filter(_filter = None):
            def __filter(node: Node) -> bool:
                return _filter(node)
            return __filter

        for node in PreOrderIter(self._root, __decorate_filter(_filter)):
            yield node.obj

    # simple wrapper for anytree
    def findall(self, filter_=None):
        from anytree import findall as anytree_findall
        found = anytree_findall(self._root, filter_)
        objects = list()
        for f in found:
            objects.append(f.obj)
        return objects

    # XXX puke, need to add some logging, but keeping the print template for now
    def add(self, name, obj, typ, **kwargs) -> Node:
        #print('DEBUG {}'.format(type(obj)))
        #print('DEBUG {} is a {} {}'.format(type(obj), meters.Meter,
        #                                   issubclass(type(obj), meters.Meter)))
        #print('DEBUG {} is a {} {}'.format(type(obj), meters.MeterInterface,
        #                                   issubclass(type(obj), meters.MeterInterface)))
        #print('DEBUG {} is a {} {}'.format(type(obj), devices.DeviceInterface,
        #                                   issubclass(type(obj), devices.DeviceInterface)))
        #print('DEBUG {} has a {} {}'.format(type(obj), '_device', hasattr(obj, '_device')))
        #print('DEBUG {} has a {} {}'.format(type(obj), devices.VirtualDevice,
        #                                    #issubclass(type(obj._device), devices.VirtualDevice)))
        #                                    isinstance(obj._device, devices.VirtualDevice)))
        #print('DEBUG {} has a {} {}'.format(type(obj), devices.Bmp390,
        #                                    #issubclass(type(obj._device), devices.Bmp390)))
        #                                    isinstance(obj._device, devices.Bmp390)))

        # XXX right now only supporting virtual or i2c, needs more work when bus is 
        # abstracted away and there are multiple physical buses
        if hasattr(obj, '_device') is False:
            raise ValueError('cannot determine if {} is physical or virtual'.format(obj))
        if isinstance(obj._device, devices.VirtualDevice):
            parent = self._virtual_bus
            typ = constants.VIRTUAL
        else:
            parent = self._i2c_bus
            typ = NONE

        if issubclass(type(obj), meters.MeterInterface):
            typ = typ | constants.METER
        elif issubclass(type(obj), controls.MuxInterface):
            typ = typ | constants.MUX
        elif issubclass(type(obj), devices.DeviceInterface):
            typ = typ | constants.DEVICE
        else:
            m = 'unsupported type, must be one of MeterInterface, MuxInterface, or DeviceInterface'
            raise ValueError(m)

        #print('DEBUG is_virtual {}'.format(bool(typ & constants.VIRTUAL)))
        #print('DEBUG is_meter {}'.format(bool(typ & constants.METER)))
        #print('DEBUG is_mux {}'.format(bool(typ & constants.MUX)))
        #print('DEBUG is_device {}'.format(bool(typ & constants.DEVICE)))

        n = Node(name, parent=self._virtual_bus, obj=obj, metadata=Metadata(typ), **kwargs)
        return n

    def _build_tree(self, i2c, parent, store: datastructures.Store,
                    env: dict[str, Any] | None = None, addr_filter: set = set()):
        try:
            if i2c.try_lock():
                devs = i2c.scan()
            i2c.unlock()
        except AttributeError:
            devs = i2c.scan()

        logger.debug('initial scan results: %s, applying filter: %s',
                     [hex(n) for n in devs],
                     [hex(n) for n in addr_filter])

        devs = [n for n in devs if n not in addr_filter]

        for addr in devs:
            logger.info('building node for address: %s', hex(addr))

            try:
                d = profiles.profiles[addr]

                self._build_node(i2c, addr, d, parent, store, env)
            except KeyError as e:
                message = 'bus scan reports unsupported address {}, continuing...'.format(hex(addr))
                logger.warning(message)

    def _build_node(self, i2c, address, profile, parent, store: datastructures.Store,
                    env: dict[str, Any] | None = None):
        if profile.is_mux():
            logger.info('multiplexer %s found at addr %s, setting up multiple channels',
                        profile.name, hex(address))

            mux = controls.mux_factory.get_mux(i2c, profile.name, profile.device_id,
                                               profile.capabilities, address)
            logger.info('multiplexer supported channels: %s', len(mux))

            node = Node(address, parent=parent, obj=mux,
                        metadata=Metadata(constants.MUX))
            
            def channel_bus(channel) -> int:
                r = 0
                if 'channel_switch' in channel.__dict__:
                    n = int.from_bytes(channel.channel_switch, 'little')
                    while n > 1:
                        n = n >> 1
                        r = r + 1
                return r
            for channel in mux.channels():
                chan = Node(channel_bus(channel), parent=node, obj=channel,
                            metadata=Metadata(constants.CHANNEL))
                addr_set = set()
                addr_set.add(mux.address)
                self._build_tree(channel, chan, store, env, addr_set)
        else:
            dev = devices.device_factory.get_device(i2c, profile.name, profile.device_id,
                                                    profile.capabilities, address, env)
            node = Node(address, parent=parent, obj=dev,
                        metadata=Metadata(constants.DEVICE))
            self._build_leaves(i2c, address, dev, node, store)

    def _build_leaves(self, i2c, address, device, parent, store: datastructures.Store):
        # XXX only do this for meters, eventually need to discriminate with detectors
        for cap in device.capabilities_gen():
            try:
                m = meters.meter_factory.get_meter(cap, device, store)
                leaf = Node(constants.to_capability_strings[cap], parent=parent, obj=m,
                            metadata=Metadata(constants.METER))
            except ValueError as e:
                logger.warning('name %s, board %s, capability %s - no ctor',
                               device.name, device.board, constants.to_capability_strings[cap])



import abc
from collections.abc import Iterator
import logging
import typing

from busio import I2C

from . import datastructures
from . import devices
from .constants import *

logger = logging.getLogger(__name__)

class MeterInterface(devices.DeviceInterface, metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'measure') and
                callable(subclass.measure) and
                hasattr(subclass, 'measurement') and
                callable(subclass.measurement) and
                hasattr(subclass, 'units') and
                callable(subclass.units) or
                NotImplemented)

    @abc.abstractmethod
    def measure(self) -> float:
        raise NotImplementedError

    @abc.abstractmethod
    def measurement(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def units(self) -> str:
        raise NotImplementedError

class Meter(MeterInterface):
    def __init__(self, device):
        self._device = device

    @property
    def address(self) -> int:
        return self._device.address

    @property
    def board(self) -> int:
        return self._device.board

    @property
    def name(self) -> str:
        return self._device.name

    @property
    def capabilities(self) -> list[int]:
        return self._device.capabilities

    @property
    def capabilities_gen(self) -> Iterator[int]:
        for cap in self._device.capabilities:
            yield cap

    @property
    def bus(self) -> I2C:
        return self._device.bus

    @property
    def bus_id(self) -> int:
        return self._device.bus_id

#
# BMP390 based measurements
class Bmp390Temperature(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        bmp = self._real_device
        return bmp.temperature

    @property
    def measurement(self) -> int:
        return TEMPERATURE

    @property
    def units(self) -> str:
        return CELSIUS_UNITS

    @property
    def real_device(self):
        return self._real_device

class Bmp390Pressure(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        bmp = self._real_device
        return bmp.pressure

    @property
    def measurement(self) -> int:
        return PRESSURE

    @property
    def units(self) -> str:
        return HECTOPASCAL_UNITS

    @property
    def real_device(self):
        return self._real_device

class Bmp390Altitude(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        bmp = self._real_device
        return bmp.altitude

    @property
    def measurement(self) -> int:
        return ALTITUDE

    @property
    def units(self) -> str:
        return METER_UNITS

    @property
    def real_device(self):
        return self._real_device
# END BMP390

#
# SHT41 (SHT4x) measurements
class Sht41Temperature(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        sht = self._real_device
        temp, _ =  sht.measurements
        return temp

    @property
    def measurement(self) -> int:
        return TEMPERATURE

    @property
    def units(self) -> str:
        return CELSIUS_UNITS

    @property
    def real_device(self):
        return self._real_device

class Sht41RelativeHumidity(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        sht = self._real_device
        _, rh = sht.measurements
        return rh

    @property
    def measurement(self) -> int:
        return RELATIVE_HUMIDITY

    @property
    def units(self) -> str:
        return PERC_RELATIVE_HUMIDITY_UNITS

    @property
    def real_device(self):
        return self._real_device
# END SHT41 (SHT4x)

#
# VEML7700 measurements
class Veml7700AmbientLight(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        veml7700 = self._real_device
        l = veml7700.light
        return l

    @property
    def measurement(self) -> int:
        return AMBIENT_LIGHT

    @property
    def units(self) -> str:
        return AMBIENT_LIGHT_UNITS

    @property
    def real_device(self):
        return self._real_device

class Veml7700Lux(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        veml7700 = self._real_device
        l = veml7700.lux
        return l

    @property
    def measurement(self) -> int:
        return LUX

    @property
    def units(self) -> str:
        return LUX_UNITS

    @property
    def real_device(self):
        return self._real_device
# END VEML7700

#
# SCD41 (SCD4x) measurements
class Scd41Temperature(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        scd = self._real_device
        temp = scd.temperature
        return temp

    @property
    def measurement(self) -> int:
        return TEMPERATURE

    @property
    def units(self) -> str:
        return CELSIUS_UNITS

    @property
    def real_device(self):
        return self._real_device

class Scd41RelativeHumidity(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        scd = self._real_device
        rh = scd.relative_humidity
        return rh

    @property
    def measurement(self) -> int:
        return RELATIVE_HUMIDITY

    @property
    def units(self) -> str:
        return PERC_RELATIVE_HUMIDITY_UNITS

    @property
    def real_device(self):
        return self._real_device

class Scd41CO2(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> int:
        scd = self._real_device
        co2 = scd.CO2
        return co2

    @property
    def measurement(self) -> int:
        return CO2

    @property
    def units(self) -> str:
        return PPM_UNITS

    @property
    def real_device(self):
        return self._real_device
# END SCD41 (SCD4x)

#
# TSL2591 measurements
class Tsl2591Lux(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        tsl2591 = self._real_device
        l = tsl2591.lux
        return l

    @property
    def measurement(self) -> int:
        return LUX

    @property
    def units(self) -> str:
        return LUX_UNITS

    @property
    def real_device(self):
        return self._real_device

class Tsl2591Infrared(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        tsl2591 = self._real_device
        l = tsl2591.infrared
        return l

    @property
    def measurement(self) -> int:
        return INFRARED

    @property
    def units(self) -> str:
        return AMBIENT_LIGHT_UNITS

    @property
    def real_device(self):
        return self._real_device

class Tsl2591Visible(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        tsl2591 = self._real_device
        l = tsl2591.visible
        return l

    @property
    def measurement(self) -> int:
        return VISIBLE

    @property
    def units(self) -> str:
        return AMBIENT_LIGHT_UNITS

    @property
    def real_device(self):
        return self._real_device

class Tsl2591FullSpectrum(Meter):
    def __init__(self, device, store: datastructures.Store | None = None):
        super().__init__(device)
        self._real_device = device.real_device
        self._store = store

    @property
    def measure(self) -> float:
        tsl2591 = self._real_device
        l = tsl2591.full_spectrum
        return l

    @property
    def measurement(self) -> int:
        return FULL_SPECTRUM

    @property
    def units(self) -> str:
        return AMBIENT_LIGHT_UNITS

    @property
    def real_device(self):
        return self._real_device
# END TSL2591

class MeterFactory:
    def __init__(self):
        self._boards = {}

    def register_board(self, board, ctor):
        try:
            dev_board = self._boards[board]
            dev_board['ctor'] = ctor
        except KeyError as e:
            self._board[board] = dict()
            self._board[board]['ctor'] = ctor

    def register_meter(self, board, measurement: int, ctor):
        try:
            dev_board = self._boards[board]
            dev_board[measurement] = ctor
        except KeyError as e:
            self._boards[board] = dict()
            self._boards[board][measurement] = ctor

    def get_meter(self, measurement: int, device: devices.Device,
                  store: datastructures.Store | None = None) -> MeterInterface:
        dev_board = self._boards.get(device.board)
        ctor = dev_board.get(measurement)
        if not ctor:
            raise ValueError('{}:{}'.format(device.board, measurement))

        return ctor(device, store)

meter_factory = MeterFactory()
meter_factory.register_meter(BMP390, TEMPERATURE, Bmp390Temperature)
meter_factory.register_meter(BMP390, PRESSURE, Bmp390Pressure)
meter_factory.register_meter(BMP390, ALTITUDE, Bmp390Altitude)
meter_factory.register_meter(SHT41, TEMPERATURE, Sht41Temperature)
meter_factory.register_meter(SHT41, RELATIVE_HUMIDITY, Sht41RelativeHumidity)
meter_factory.register_meter(VEML7700, AMBIENT_LIGHT, Veml7700AmbientLight)
meter_factory.register_meter(VEML7700, LUX, Veml7700Lux)
meter_factory.register_meter(SCD41, TEMPERATURE, Scd41Temperature)
meter_factory.register_meter(SCD41, RELATIVE_HUMIDITY, Scd41RelativeHumidity)
meter_factory.register_meter(SCD41, CO2, Scd41CO2)
meter_factory.register_meter(TSL2591, LUX, Tsl2591Lux)
meter_factory.register_meter(TSL2591, INFRARED, Tsl2591Infrared)
meter_factory.register_meter(TSL2591, VISIBLE, Tsl2591Visible)
meter_factory.register_meter(TSL2591, FULL_SPECTRUM, Tsl2591FullSpectrum)



from importlib import import_module
import logging
from typing import Any

from anytree import LevelOrderIter
from busio import I2C

from .calibration import Calibration
from .config import Config
from .constants import (to_device_type,
                        to_device_ids,
                        VIRTUAL,
)
from .datastructures import Store
from .devices import device_factory, DeviceInterface
from .devicetree import DeviceTree
from .tools.mixins import RunnableMixin, SchedulableMixin

logger = logging.getLogger(__name__)

class SensorKit(RunnableMixin):
    def __init__(self, bus: I2C, config: dict[str, Any], scheduler,
                 store: dict[str, Any] | None = None):
        self._bus = bus
        self._config = Config(config)

        self._env = self._config.env

        self._store = Store(store)
        self._store.tree = DeviceTree(bus, self._store, self._env)
        self._store.tree.build()
        self._scheduler = scheduler

        self._calibrations = []

        self._static_args = {
            'store': self._store,
            'scheduler': self._scheduler,
        }

        self._virtual_devices = self._config.virtual_devices
        for dev in self._virtual_devices:
            conf = self._virtual_devices[dev]
            objs = self._instantiate_device(conf)

            for d in objs:
                self._store.tree.add(dev, d, (to_device_type[conf['type']] | VIRTUAL))

        calibrations = self._config.calibrations
        self._build_calibrations(calibrations)

    @property
    def tree(self) -> DeviceTree:
        return self._store.tree

    def run(self):
        # Order:
        #   Pre:
        #     - SchedulableMixins
        #
        #   Run:
        #     - Runnables
        for cobj in self._calibrations:
            if isinstance(cobj, SchedulableMixin):
                cobj.schedule(True)

        for node in LevelOrderIter(self._store.tree._root):
            if node.obj is not None and isinstance(node.obj, RunnableMixin):
                node.obj.run()

    def stop(self):
        # Order:
        #   Pre:
        #     - SchedulableMixins
        #
        #   Run:
        #     - Runnables
        for cobj in self._calibrations:
            if isinstance(cobj, SchedulableMixin):
                cobj.unschedule()

        for node in LevelOrderIter(self._store.tree._root):
            if node.obj is not None and isinstance(node.obj, RunnableMixin):
                node.obj.stop()

    def _build_calibrations(self, calibrations) -> None:
        for c in calibrations:
            for d in self._store.tree.devices_iter(lambda node: node.obj.board == to_device_ids[c]):
                for conf in calibrations[c]:
                    cobj = Calibration(c, conf, d, self._store, self._scheduler)
                    self._calibrations.append(cobj)

    def _instantiate_device(self, config: dict[str, Any]) -> DeviceInterface:
        module = import_module(config['module'], package='sensorkit')

        builder_name = config['builder']
        builder = getattr(module, builder_name)
        build_obj = builder(config['capabilities'])

        args = { **self._static_args, **config['args'] }
        objects = build_obj(**args)

        return objects
